import numpy as np
from matplotlib import pyplot as plt
from fwk.coloring import ccolors

import dart.pyVII.vUtils as blxU
import sys
#import dart.pyVII.vStructures as vStruct

# 3D spline with Bézier curves
from scipy.interpolate import bisplrep
from scipy.interpolate import bisplev
from scipy.interpolate import splrep
from scipy.interpolate import splev

from scipy.interpolate import interp1d

from scipy.interpolate import RBFInterpolator

import fwk


class Interpolator:
    def __init__(self, _dartAPI, _blw, _cfg):
        
        self.iSolver = _dartAPI
        self.blw = _blw

        # Create data structures for interpolation
        self.tms = fwk.Timers()
        self.tms['GetWing'].start()
        self.__GetWing(_blw, _cfg)
        self.tms['GetWing'].stop()

        self.cfg = _cfg
        print(self.tms)

        self.stgPt=np.zeros(self.cfg['nSections'], dtype=int)

        self.xxPrev = [[np.zeros(0) for iReg in range(3)] for iSec in range(self.cfg['nSections'])]
        #self.uePrev = [[np.zeros(len(self.viscStruct[iSec][iReg][:,0])) for iReg in range(3)] for iSec in range(len(self.viscStruct))]
        self.DeltaStarPrev = [[np.zeros(0) for iReg in range(3)] for iSec in range(self.cfg['nSections'])]


    def GetInvBC(self, vSolver, couplIter):
        self.tms['InvToVisc'].start()

        # Extract parameters from the inviscid solver
        # Particular care must be taken with space directions; In the inviscid, the airfoil is in the x,z plane.
        # While setting up the viscous solver, y and z must be inverted. 

        # Wing

        U = [np.zeros((len(self.invStruct[i][:,0]), 3)) for i in range(len(self.invStruct))]
        M = [np.zeros(len(self.invStruct[i][:,0])) for i in range(len(self.invStruct))]
        Rho = [np.zeros(len(self.invStruct[i][:,0])) for i in range(len(self.invStruct))]

        # Extract variables.

        for iReg in range(len(self.invStruct)):
            for iNode, row in enumerate(self.invStruct[iReg][:,3]):
                row = int(row)
                U[iReg][iNode,:] = [self.iSolver.U[row][0], self.iSolver.U[row][1], self.iSolver.U[row][2]]
                M[iReg][iNode] = self.iSolver.M[row]
                Rho[iReg][iNode] = self.iSolver.rho[row]
        
            if self.cfg['nDim'] == 3:
                U[iReg][:,[1,2]] = U[iReg][:,[2,1]]
        
        # Find stagnation point
        nD = self.cfg['nDim']
        for iSec in range(len(self.viscStruct_tr)):
            Ux = [self.__RbfToSections(self.invStruct_tr[0][:,:nD], U[0][:,0], self.viscStruct_tr[iSec][0][:,:nD]), self.__RbfToSections(self.invStruct_tr[1][:, [0,2] if nD==3 else 0], U[1][:,0], self.viscStruct[iSec][1][:,[0,2] if nD==3 else 0])]
            Uy = [self.__RbfToSections(self.invStruct_tr[0][:,:nD], U[0][:,1], self.viscStruct_tr[iSec][0][:,:nD]), self.__RbfToSections(self.invStruct_tr[1][:, [0,2] if nD==3 else 0], U[1][:,1], self.viscStruct[iSec][1][:,[0,2] if nD==3 else 0])]
            Uz = [self.__RbfToSections(self.invStruct_tr[0][:,:nD], U[0][:,2], self.viscStruct_tr[iSec][0][:,:nD]), self.__RbfToSections(self.invStruct_tr[1][:, [0,2] if nD==3 else 0], U[1][:,2], self.viscStruct[iSec][1][:,[0,2] if nD==3 else 0])]
            Me = [self.__RbfToSections(self.invStruct_tr[0][:,:nD], M[0], self.viscStruct_tr[iSec][0][:,:nD]), self.__RbfToSections(self.invStruct_tr[1][:, [0,2] if nD==3 else 0], M[1], self.viscStruct[iSec][1][:,[0,2] if nD==3 else 0])]
            Rhoe = [self.__RbfToSections(self.invStruct_tr[0][:,:nD], Rho[0], self.viscStruct_tr[iSec][0][:,:nD]), self.__RbfToSections(self.invStruct_tr[1][:, [0,2] if nD==3 else 0], Rho[1], self.viscStruct[iSec][1][:,[0,2] if nD==3 else 0])]
            
            if couplIter == 0:
                self.stgPt[iSec] = np.argmin(Ux[0]**2+Uy[0]**2)
            
                xUp, xLw = self.__Remesh(self.viscStruct[iSec][0][:,0], self.stgPt[iSec])
                yUp, yLw = self.__Remesh(self.viscStruct[iSec][0][:,1], self.stgPt[iSec])
                zUp, zLw = self.__Remesh(self.viscStruct[iSec][0][:,2], self.stgPt[iSec])

                vSolver.SetGlobMesh(iSec, 0, xUp, yUp, zUp)
                vSolver.SetGlobMesh(iSec, 1, xLw, yLw, zLw)
                vSolver.SetGlobMesh(iSec, 2, self.viscStruct[iSec][1][:,0],
                                            self.viscStruct[iSec][1][:,1],
                                            self.viscStruct[iSec][1][:,2])

                self.DeltaStarPrev[iSec][0] = np.zeros(len(xUp))
                self.DeltaStarPrev[iSec][1] = np.zeros(len(xLw))
                self.DeltaStarPrev[iSec][2] = np.zeros(len(self.viscStruct[iSec][1][:,0]))
                
                self.xxPrev[iSec][0] = np.zeros(len(xUp))
                self.xxPrev[iSec][1] = np.zeros(len(xLw))
                self.xxPrev[iSec][2] = np.zeros(len(self.viscStruct[iSec][1][:,0]))

            UxUp, UxLw = self.__Remesh(Ux[0], self.stgPt[iSec])
            UyUp, UyLw = self.__Remesh(Uy[0], self.stgPt[iSec])
            UzUp, UzLw = self.__Remesh(Uz[0], self.stgPt[iSec])

            plt.plot(xUp, UxUp)
            plt.plot(xLw, UxLw)
            plt.show()

            vSolver.SetVelocities(iSec, 0, UxUp, UyUp, UzUp)
            vSolver.SetVelocities(iSec, 1, UxLw, UyLw, UzLw)
            vSolver.SetVelocities(iSec, 2, Ux[1], Uy[1], Uz[1])

            MeUp, MeLw = self.__Remesh(Me[0], self.stgPt[iSec])

            vSolver.SetMe(iSec, 0, MeUp)
            vSolver.SetMe(iSec, 1, MeLw)
            vSolver.SetMe(iSec, 2, Me[1])

            RhoUp, RhoLw = self.__Remesh(Rhoe[0], self.stgPt[iSec])

            vSolver.SetRhoe(iSec, 0, RhoUp)
            vSolver.SetRhoe(iSec, 1, RhoLw)
            vSolver.SetRhoe(iSec, 2, Rhoe[1])

            for iReg in range(3):
                vSolver.SetDeltaStarExt(iSec, iReg, self.DeltaStarPrev[iSec][iReg])
                vSolver.SetxxExt(iSec, iReg, self.xxPrev[iSec][iReg])
            
            

    def SetBlowingVel(self, vSolver, couplIter):

        for iSec in range(len(self.viscStruct)):
            for iReg in range(3):
                self.DeltaStarPrev[iSec][iReg] = vSolver.ExtractDeltaStar(iSec, iReg)
                self.xxPrev[iSec][iReg] = vSolver.Extractxx(iSec, iReg)

        x = [np.zeros((0,3)) for _ in range(2)]
        blw = [np.zeros(0) for _ in range(2)]

        for iSec in range(len(self.viscStruct)):
            for iReg in range(2):
                if iReg == 0:
                    x[iReg] = np.vstack((x[iReg], self.viscCg_tr[iSec][iReg][:,:3]))
                    blwUp = vSolver.ExtractBlowingVel(iSec, 0)
                    blwLw = vSolver.ExtractBlowingVel(iSec, 1)
                    blw[iReg] = np.concatenate((blw[iReg], blwUp[::-1] , blwLw))
                elif iReg == 1:
                    x[iReg] = np.vstack((x[iReg], self.viscCg_tr[iSec][iReg][:,:3]))
                    blw[iReg] = np.concatenate((blw[iReg], vSolver.ExtractBlowingVel(iSec, 2)))

        self.tms['Interpolation'].start()
        nD = self.cfg['nDim']
        mappedBlw = [self.__RbfToSections(x[0], blw[0], self.invCg_tr[0][:,:3]), self.__RbfToSections(x[1][:, [0,2] if nD==3 else 0], blw[1], self.invCg_tr[1][:, [0,2] if nD==3 else 0])]
        
        self.tms['Interpolation'].stop()
        for iElm in range(len(self.invCg_tr[0][:,0])):
            self.blw[0].setU(iElm, mappedBlw[0][iElm])
        for iElm in range(len(self.invCg_tr[1][:,0])):
            self.blw[1].setU(iElm, mappedBlw[1][iElm])

    def RunSolver(self):
        self.iSolver.run()

    def ResetSolver(self):
        self.iSolver.reset()

    def GetCp(self,bnd):
        import dart.utils as floU
        Cp = floU.extract(bnd, self.iSolver.Cp)
        return Cp

    def GetAlpha(self):
        return self.iSolver.pbl.alpha

    def GetMinf(self):
        return self.iSolver.pbl.M_inf
    
    def GetCl(self):
        return self.iSolver.Cl
    
    def GetCm(self):
        return self.iSolver.Cm

    def __Remesh(self, vec, stgPt):
        xUp = vec[:stgPt+1]
        xUp = xUp[::-1]

        xLw = vec[stgPt:]
        return xUp, xLw

    def __RbfToSections(self, x, var, xs):
        if np.all(var == var[0]):
            varOut = RBFInterpolator(x, var, neighbors=1, kernel='linear', smoothing=0.0, degree=0)(xs)
        else:
            varOut = RBFInterpolator(x, var, neighbors=self.cfg['neighbors'], kernel=self.cfg['rbftype'], smoothing=self.cfg['smoothing'], degree=self.cfg['degree'])(xs)
        return varOut


    def __GetWing(self, blw, config):

        viscStruct = [[np.zeros((config['nPoints'][1] if i==1 else 2*config['nPoints'][0]-1, 3)) for i in range(2)] for _ in range(config['nSections'])]
        viscStruct_tr = [[np.zeros((config['nPoints'][1] if i==1 else 2*config['nPoints'][0]-1, 3)) for i in range(2)] for _ in range(config['nSections'])]

        
        invStruct = [np.zeros((len(blw[i].nodes), 4)) for i in range(len(blw))]
        invStruct_tr = [np.zeros((len(blw[i].nodes), 4)) for i in range(len(blw))]

        # Wing
        upperNodes = np.zeros((0,4))
        lowerNodes = np.zeros((0,4))

        self.tms['ElmLoop'].start()
        # Compute reference normal
        nVec = [np.zeros(0) for _ in range(blw[1].tag.elems[0].nodes.size())]
        for iNode, n in enumerate(blw[1].tag.elems[0].nodes):
            nVec[iNode] = np.array([n.pos[0], n.pos[1], n.pos[2]])
        normalRef = np.cross(nVec[1] - nVec[0], nVec[2] - nVec[0])
        normalRef/=np.linalg.norm(normalRef)
            
        for e in blw[0].tag.elems:
            nVec = [np.zeros(0) for _ in range(e.nodes.size())]
            for iNode, n in enumerate(e.nodes):
                nVec[iNode] = np.array([n.pos[0], n.pos[1], n.pos[2]])
            normal = np.cross(nVec[1] - nVec[0], nVec[2] - nVec[0])
            normal/=np.linalg.norm(normalRef)
            if np.dot(normal, normalRef) >= 0:
                for n in e.nodes:
                    if n.row not in upperNodes[:,3] and n.row not in lowerNodes[:,3]:
                        upperNodes = np.vstack((upperNodes, [n.pos[0], n.pos[1], n.pos[2], n.row]))
            if np.dot(normal, normalRef) < 0:
                for n in e.nodes:
                    if n.row not in lowerNodes[:,3] and n.row not in upperNodes[:,3]:
                        lowerNodes = np.vstack((lowerNodes, [n.pos[0], n.pos[1], n.pos[2], n.row]))

        self.tms['ElmLoop'].stop()

        dummyLower = lowerNodes.copy()
        dummyLower[:,0] *= -1

        invStruct_tr[0] = np.vstack((upperNodes, dummyLower))
        invStruct[0] = np.vstack((upperNodes, lowerNodes))

        # Remove duplicates
        #invStruct_tr[0] = np.unique(invStruct_tr[0], axis=0)

        # Wake part
        for iNode, n in enumerate(blw[1].nodes):
            invStruct_tr[1][iNode, :] = [n.pos[0], n.pos[1], n.pos[2], n.row]
        invStruct[1] = invStruct_tr[1].copy()

        # Create spanwise distribution
        if config['nDim'] == 2:
            spanDistri = np.zeros(1)
            spanDistri[0] = invStruct[0][0,2] # Span direction is the z direction

            wingSpan = 0.
            fInterp = interp1d(invStruct_tr[0][:,0], invStruct_tr[0][:,1])

        elif config['nDim'] == 3:
            invStruct[0][:, [1,2]] = invStruct[0][:, [2, 1]]
            invStruct[1][:, [1,2]] = invStruct[1][:, [2, 1]]
            invStruct_tr[0][:, [1,2]] = invStruct_tr[0][:, [2, 1]]
            invStruct_tr[1][:, [1,2]] = invStruct_tr[1][:, [2, 1]]
            spanDistri = np.linspace(min(invStruct[0][:,2]), max(invStruct[0][:,2]), config['nSections'])
            spanDistri[0] +=0.1
            spanDistri[-1] -=0.1

            # Spline wing
            # Point in the wing tip region are not accounted for during Bezier curve spline

            m = len(invStruct_tr[0][:,0])
            _s = (m-np.sqrt(2*m) + m+np.sqrt(2*m))/2
            tck_wing = bisplrep(invStruct_tr[0][invStruct_tr[0][:,2]<config['span'],0], invStruct_tr[0][invStruct_tr[0][:,2]<config['span'],2], invStruct_tr[0][invStruct_tr[0][:,2]<config['span'],1], kx=5, ky=5, s=1e-6, quiet=1)

        # Create sections
        self.tms['iSecLoop'].start()

        for iSec, z in enumerate(spanDistri):
            portion = invStruct[0][abs(invStruct[0][:,2] - z) <= 0.01 * config['span'],:]
            
            chordLength = max(portion[:, 0]) - min(portion[:, 0])
            xDistri = min(portion[:,0]) + 1/2*(1 + np.cos(np.linspace(0, np.pi, config['nPoints'][0])))*chordLength

            viscStruct[iSec][0][:,0] = np.concatenate((xDistri[:-1], xDistri[::-1]), axis=None)
            viscStruct[iSec][0][:,2] = z*np.ones(len(viscStruct[iSec][0][:,0]))

            viscStruct_tr[iSec][0][:,0] = np.concatenate((xDistri[:-1], -xDistri[::-1]), axis=None).copy()
            viscStruct_tr[iSec][0][:,2] = z*np.ones(len(viscStruct[iSec][0][:,0]))

            # Interpolation in the transformed space.
            if config['nDim'] == 3:
                    
                for iPoint in range(len(viscStruct_tr[iSec][0][:,0])):
                    viscStruct_tr[iSec][0][iPoint, 1] = bisplev(viscStruct_tr[iSec][0][iPoint,0], viscStruct_tr[iSec][0][iPoint,2], tck_wing)


            elif config['nDim'] == 2:
                viscStruct_tr[iSec][0][:,1] = fInterp(viscStruct_tr[iSec][0][:,0])

            #viscStruct_tr[iSec][0][0,1] = 0.
            #viscStruct_tr[iSec][0][-1,1] = 0.
            viscStruct[iSec][0][:,1] = viscStruct_tr[iSec][0][:,1].copy()

            # Wake 
            wakeLength = np.max(invStruct[1][:,0]) - np.max(viscStruct[iSec][0][:,0])
            viscStruct_tr[iSec][1][:,0] = np.max(viscStruct[iSec][0][:,0]) + (np.max(viscStruct[iSec][0][:,0]) + np.cos(np.linspace(np.pi, np.pi/2, config['nPoints'][1]))) * wakeLength
            viscStruct_tr[iSec][1][:,2] = z*np.ones(len(viscStruct[iSec][1][:,0]))
            viscStruct_tr[iSec][1][:,1] = interp1d(invStruct_tr[1][:,0], invStruct_tr[1][:,1])(viscStruct_tr[iSec][1][:,0])

            viscStruct[iSec][1] = viscStruct_tr[iSec][1].copy()

        # Cg positions
        self.tms['iSecLoop'].stop()

        """fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
        ax.scatter(invStruct_tr[0][:,0], invStruct_tr[0][:,2], invStruct_tr[0][:,1], s=0.3)
        ax.scatter(invStruct_tr[1][:,0], invStruct_tr[1][:,2], invStruct_tr[1][:,1], s=0.3)
        for sec in viscStruct_tr:
            ax.scatter(sec[0][:,0], sec[0][:,2], sec[0][:,1], color = 'red')
            ax.scatter(sec[1][:,0], sec[1][:,2], sec[1][:,1], color = 'red')
        ax.set_zlim([-0.5, 0.5])
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        plt.show()"""

        viscCg_tr = [[np.zeros((len(viscStruct[iSec][iReg][:,0]) - 1, 3)) for iReg in range(2)] for iSec in range(len(viscStruct))]
        invCg_tr = [np.zeros((len(blw[iReg].tag.elems), 3)) for iReg in range(2)]
        self.tms['CgLoop'].start()
        
        # Viscous
        for iSec in range(len(viscStruct)):
            for iReg in range(len(viscStruct[iSec])):
                for iElm in range(len(viscStruct[iSec][iReg][:,0]) - 1):
                    for iDir in range(3):
                        viscCg_tr[iSec][iReg][iElm, iDir] = viscStruct_tr[iSec][iReg][iElm, iDir] + viscStruct_tr[iSec][iReg][iElm + 1, iDir]
                viscCg_tr[iSec][iReg] /= 2

        # Inviscid
        for iReg in range(len(blw)):
            for iElm, e in enumerate(blw[iReg].tag.elems):
                for n in e.nodes:
                    if n.row in upperNodes[:,3] or iReg == 1:
                        invCg_tr[iReg][iElm, 0] +=n.pos[0]
                    elif n.row in lowerNodes[:,3]:
                        invCg_tr[iReg][iElm, 0] += -n.pos[0]

                    invCg_tr[iReg][iElm, 1] +=n.pos[1]
                    invCg_tr[iReg][iElm, 2] +=n.pos[2]
                invCg_tr[iReg][iElm, :] /= e.nodes.size()

            if config['nDim'] == 3:
                invCg_tr[iReg][:,[1,2]] = invCg_tr[iReg][:,[2,1]]
        self.tms['CgLoop'].stop()

        self.invStruct = invStruct
        self.invStruct_tr = invStruct_tr
        self.viscStruct = viscStruct
        self.viscStruct_tr = viscStruct_tr

        self.invCg_tr = invCg_tr
        self.viscCg_tr = viscCg_tr
        np.set_printoptions(threshold=sys.maxsize)
        for sec in self.viscStruct:
            print('next')
            sec[0][:,[1,2]] = sec[0][:,[2,1]]
            print(sec[0])
